#define INHERIT MODULE
#include "brain.h"

void module_init()
{
#ifdef SHOUTCURSE
  this_shell()->add_command("Curse %*@O","curse");
#endif
  this_shell()->add_command("frog %*@O","frog");
  this_shell()->add_command("callinfo %s","callinfo");
  this_shell()->add_command("goin %-m %R","goin");
  this_shell()->add_command("Echoto %*O %S","tell");
  this_shell()->add_command("findfunction %@*O %S","findfunction");
  this_shell()->add_command("GO %s","GO");
  this_shell()->add_command("I %-rnwlvdfsx %~O","inventory");
}

short() { return "The default shell module"; }
long() { return 
      " I [-rnwlvdfsx] <object>:Shows inventory of an object.\n"+
      " Echoto <object> <what> :Echo a message to one or more objects.\n"+
      " callinfo <name/arg>    :Checks current call_outs.\n"+
      " frog <player>          :Turn <player> frog/human.\n"+
      " goin [-m] <object>     :Go into an object.\n"+
#ifdef SHOUTCURSE
      " Curse <player>         :Give/Remove a shout curse to <player>.\n"+
#endif
      " findfunction <ob> <fun>:Tells you in what file a fun is defined.\n"+
      " GO <direction>         :Walk and try to pass obstacles.\n"+
	  "" ; }

string *module_help()
{
  return ({
      "I",":I [-rnwlvdfs] <object>\n"+
      "This will show the inventory of an object. "+
      " Default object is yourself. Flags:\n"+
      " -r : shows inventory recursivly\n"+
      " -n : toggle inventory-numbers (default is on)\n"+
      " -w : toggle showing the weight of the objects (default is off)\n"+
      " -l : toggle showing of levels/ac/wc (default is off)\n"+
      " -v : toggle value (default is off)\n"+
      " -d : toggle description (default is on)\n"+
      " -s : toggle showing any shadows (default is off)\n"+
      " -x : toggle showing extra look\n"+
      "A good idea might be to select the flags you want and make "+
      "an alias like 'alias ii I -wlv' for instance.",

    "GO",":GO <direction>\n"+
      "This commands tries to take you one step in the direction "+
      "<direction> without bothering with things like guards, locked doors, "+
      "etc. etc.",
    "findfunction",":findfunction <object(s)> <function>\n"+
      "Writes the filename in which the function is defined.",

      "Echoto",":Echoto <object(s)> <message>\n"+
      "Echoto echos a message to one of more objects. The message itself "+
      "will also be evaluated so if you want to make sure that it is "+
      "delivered as a string then quote it with \". Read 'shellhelp objects' "+
      "to find out how the shell finds the objects to deliver the message to.",

#ifdef SHOUTCURSE
      "Curse",":Curse <player>\n"+
      "Give a shout curse to a player or "+
      "remove it if he's already got one.",
#endif
      "frog",":frog <player>\n"+
      "Simply turn a player frog/human.",

      "callinfo",":callinfo <wizard/arg>\n"+
      "Callinfo prints some information about the current call_out list. "+
      "'callinfo <wizard>' Will show all call_outs currently waiting, "+
      "which were initiated by any of <wizards> objects. "+
      "'callinfo all' Will show all call_outs currently waiting. "+
      "'callinfo 0' Will show all call_outs started by objects "+
      "in /local /obj /room /open etc. "+
      "'callinfo <object>' Will show all call_outs from <object>. "+
      "'callinfo' will show your call_outs. "+
      "'callinfo statistics' will show some statistics about the current "+
      "call_outs.",

      "goin",":goin [-m] <object>\n"+
      "Go into an object. "+
      "If '-m' is used goin will move you into object2 even if you are already "+
      "there. This flag will always move you silently.",
    });
}

#define COL_RECURSIVE 1
#define COL_NUMBERS 2
#define COL_WEIGHT 4
#define COL_LEVEL 8
#define COL_VALUE 16
#define COL_SHORT 32
#define COL_FNAM 64
#define COL_SHADOW 128
#define COL_XLOOK 256

string fiddle_obj_desc(object o,int flags,int n)
{
  string tmp;
  mixed a;

  tmp="";
  if(flags & COL_NUMBERS)
    tmp+=right_format(n+"",2)+":";

  if(flags & COL_WEIGHT)
  {
    n=(int)o->query_weight();
    if(n>999)
    {
      tmp+="BIG,";
    }else{
      tmp+=right_format(o->query_weight()+"",3)+",";
    }
  }

  if(flags & COL_LEVEL)
  {
    n=(int)o->query_level();
    if(!n) n=(int)o->weapon_class();
    if(!n) n=(int)o->armour_class();
    if(n)
    {
      if(n>999)
      {
	tmp+="BIG,";
      }else{
	tmp+=right_format(n+"",3)+",";
      }
    }else{
      tmp+="  -,";
    }
  }

  if(flags & COL_VALUE)
  {
    n=(int)o->query_value();

    if(n>9999)
    {
      tmp+=" BIG,";
    }else{
      tmp+=right_format(n+"",4)+",";
    }
  }

  if(flags & COL_SHORT)
  {
    if(!(a=(string)o->short()))
    {
      if(a=NAME(o))
      {
	a+=" (invisible)";
      }else{
	a="Invisible object";
      }
    }
    if(a=="Someone (invisible)")
      a=(string)call_other(o,REAL_NAME_FUNC)+" (invisible)";
    if(flags>=COL_FNAM)
      tmp+=left_format(a,33)+",";
    else
      tmp+=a+",";
  }

  if(flags & COL_FNAM)
  {
    a=short_file_name(o);
    if(flags>COL_SHADOW)
      tmp+=left_format(a,35)+" ";
    else
      tmp+=a+" ";
  }

  if(flags & COL_SHADOW)
  {
    a=shadow(o,0);
    if(a)
      a="Shadow: "+short_file_name(a);
    else
      a="";
    tmp+=left_format(a,33);
  }

  if(flags & COL_XLOOK)
    if(a=(string)o->extra_look()) tmp+=a;

  for(n=strlen(tmp)-1;tmp[n]==' ' || tmp[n]==',';n--);
  return tmp[0..n];
}


void show_inventory(object obj,int flags,int level)
{
  int e;
  object *o;
  o=all_inventory(obj);
  for(e=0;e<sizeof(o);e++)
  {
    write("                                                 "[1..level]+
	  fiddle_obj_desc(o[e],flags,e)+"\n");
    if(flags & COL_RECURSIVE)
      show_inventory(o[e],flags,level+2);
  }
}

int inventory(int *flags,object obj)
{
  int f,e;
  object *b;

  if(!obj) obj=the_bearer();
  if(!objectp(obj))
    endwrite("No such object.\n");

  write("Object :"+short_file_name(obj)+" contains:\n");
  
  f=COL_NUMBERS+COL_SHORT+COL_FNAM;

  for(e=0;e<sizeof(flags);e++)
    if(flags[e]&1) f^=COL_RECURSIVE<<e;

  show_inventory(obj,f,0);
  return 1;
}

#ifdef SHOUTCURSE
/* This (and frog) ought to take an optional argument 1 or 0 */
void curse(object obj)
{
  object b;
  
  if(!interactive(obj))
  {
    write("That is not a player.\n");
  }else{
    if(b=present("shout_curse",obj))
    {
      destruct(b);
      write("You free "+NAME(obj)+" from a shout curse.\n");
    }else{
      clone_object(SHOUTCURSE)->start(obj);
      write("You give "+NAME(obj)+" a shout curse.\n");
    }    
  }
}
#endif

void frog(object obj)
{
  if(!living(obj))
  {
    write("That is not a living.\n");
  }else{
    if(!obj->frog_curse(!obj->query_frog()))
      write("You turn "+NAME(obj)+" human again.\n");
    else
      write("You turn "+NAME(obj)+" into a frog.\n");
  }
}

int callinfo(string str)
{
  mixed *b,c,d,y,q;
  int e;
  if(!str || str=="")
    str=(string)call_other(the_bearer(),REAL_NAME_FUNC);

  if(str=="0") str=(string)0;
  b=call_out_info();

  if(str=="statistics")
  {
    c=({({}),({})});
    for(e=0;e<sizeof(b);e++)
    {
      if(b[e] && b[e][0])
      {
#if 0
	d=my_creator(b[e][0]);
#else
	d=creator(b[e][0]);
#endif
	if(!d) d="NoBody";
	y=member_array(d,c[0]);
	if(y<0)
	{
	  c=({c[0]+({d}),c[1]+({({1,b[e][2]})})});
	}else{
	  c[1][y][0]+=1;
	  c[1][y][1]+=b[e][2];
	}
      }
    }

    write("Name            Calls   Avg.len.  calls/second\n");
    for(e=0;e<sizeof(c[0]);e++)
    {
      y=c[1][e][0]*c[1][e][0]*100/(c[1][e][1]||1);
      d="000"+(y % 100);
      q=y/100+"."+extract(d,strlen(d)-2,strlen(d)-1);

      y=c[1][e][1]*100/c[1][e][0];
      d="000"+(y % 100);
      y=y/100+"."+extract(d,strlen(d)-2,strlen(d)-1);
      
      write(left_format(c[0][e],15)+
	    left_format(c[1][e][0]+"",5)+
	    right_format(y,7)+
	    right_format(q,7)+
	    "\n");
    }
  }else{
    if(stringp(str) && str[0]>='A' && str[0]<='Z')
    {
      str=lower_case(str);
    }
    else
    {
      y=(mixed)this_shell()->parse_string(str,T_STRING | T_OBJECT | T_POINTER);

      if(objectp(y))
      {
	if(interactive(y))
	{
	  str=(string)call_other(y,REAL_NAME_FUNC);
	}else{
	  str="all";
	  d=({y});
	}
      }
      else if(pointerp(y))
      {
	str="all";
	d=y;
      }
      else if(stringp(y))
      {
	str=lower_case(y);
      }
    }

    for(e=0;e<sizeof(b);e++)
    {
      if(b[e] &&
	 b[e][0] &&
	 (!d || -1!=member_array(b[e][0],d)) &&
	 (creator(b[e][0])==str || str=="all"))
      {
	write("delay:"+left_format(b[e][2]+"",5));
	write(fast_desc(b[e][0])+"->"+b[e][1]+"("+fast_desc(b[e][3])+")\n");
      }
    }
  }
  return 1;
}

int goin(int *flags,object obj)
{
  int force_move;
  force_move=flags[0];

  if(BEARERS_ENV()==obj && !force_move)
    endwrite("You are already here.\n");

  if(!flagval("muffle") && !force_move && the_bearer())
  {
    assign("back",BEARERS_ENV());
    the_bearer()->move_player("X",obj);
  }else{
    assign("back",BEARERS_ENV());
    move_object(the_bearer(),obj);
    if(the_bearer()) the_bearer()->look();
  }
  return 1;
}

private array_eq(mixed *a,mixed *b)
{
  return sizeof(a)==sizeof(b) && sizeof(a & b)==sizeof(a);
}

string tell(mixed *who,string what)
{
  int e;
  if(!pointerp(who)) who=({who});
    
  for(e=0;e<sizeof(who);e++)
    if(objectp(who[e]))
      tell_object(who[e],what+"\n");

#ifdef ECHOLOG
  if(array_eq(who,users()))
  {
    basepatch("log",ECHOLOG,"Echoto users "+what+"\n");
  }else{
    string tmp;
    tmp="";
    for(e=0;e<sizeof(who);e++)
      if(objectp(who[e]))
	tmp+=","+fast_desc(who[e]);

    tmp=tmp[1..100000];
    basepatch("log",ECHOLOG,"Echoto ({"+tmp+"}) "+what+"\n");
  }

#endif
  return "Ok.\n";
}

int findfunction(object o,string fun)
{
  string tmp;
  if(fun=="heart_beat")
    endwrite("The function 'heart_beat' is impossible to trace.\n");
  tmp=find_function(fun,o);
  if(tmp)
  {
    endwrite(fast_desc(o)+"->"+fun+" is defined in "+tmp+".\n");
  }else{
    endwrite("There is no function '"+fast_desc(o)+"->"+fun+"'.\n");
  }
}

int GO(string s)
{
  object o;
  mixed *dd,where;
  int e;
  string alt_s,a,b;
  alt_s=s;

  switch(s)
  {
  case "n": s="north"; break;
  case "s": s="south"; break;
  case "e": s="east"; break;
  case "w": s="west"; break;
  case "ne": s="northeast"; break;
  case "sw": s="southwest"; break;
  case "se": s="southeast"; break;
  case "nw": s="northwest"; break;
  case "u": s="up"; break;
  case "d": s="down"; break;
  case "o": s="out"; break;
  case "mu": s="meta-up"; break;
  case "md": s="meta-down"; break;
  case "mn": s="meta-north"; break;
  case "ms": s="meta-south"; break;
  case "me": s="meta-east"; break;
  case "mw": s="meta-west"; break;
  case "i": s="in"; break; break;

  case "north": alt_s="n"; break;
  case "south": alt_s="s"; break;
  case "east": alt_s="e"; break;
  case "west": alt_s="w"; break;
  case "northeast": alt_s="ne"; break;
  case "southwest": alt_s="sw"; break;
  case "southeast": alt_s="se"; break;
  case "northwest": alt_s="nw"; break;
  case "up": alt_s="u"; break;
  case "down": alt_s="d"; break;
  case "out": alt_s="o"; break;
  case "meta-up": alt_s="mu"; break;
  case "meta-down": alt_s="md"; break;
  case "meta-north": alt_s="mn"; break;
  case "meta-south": alt_s="ms"; break;
  case "meta-east": alt_s="me"; break;
  case "meta-west": alt_s="mw"; break;
  case "in": alt_s="i"; break; break;
  }

  o=environment(this_player());
  dd=(mixed)environment(this_player())->query_dest_dir();
#if MAPPINGS
  if(mappingp(dd))
  {
    where=dd[s];
    if(!where) where=dd[alt_s];
  }else
#endif
    if(pointerp(dd))
    {
      e=member_array(s,dd);
      if(e>0) where=dd[e-1];
      if(!where)
      {
	e=member_array(alt_s,dd);
	if(e>0) where=dd[e-1];
      }
    }
  if(!where)
  {
    dd=all_inventory(environment(this_player()));
    for(e=0;e<sizeof(dd);e++)
    {
      if(s==(string)dd[e]->query_direction() ||
	 alt_s==(string)dd[e]->query_direction() )
      {
	where=(mixed)dd[e]->query_destination();
	if(!where) where=(mixed)dd[e]->query_dest();
	if(where) break;
      }
    }
  }

  if(stringp(where) && sscanf(where,"~%s",a))
  {
    if(sscanf(a,"%s/%s",a,b))
    {
      a=WIZDIR+"/"+a+"/"+b;
    }else{
      string *dest2;
      dest2=my_explode(file_name(environment(this_player())),"/");
      dest2[sizeof(dest2)-1]=a;
      a=implode(dest2,"/");
    }
    where=a;
  }
  
  if(where)
  {
    if(objectp(where))
      where=my_file_name(where);
    if(stringp(where))
      this_player()->move_player(s+"#"+where);
  }else{
    /* last resort */
    INTERNAL_COMMAND(s);
  }
  if(o==environment(this_player()))
    write("Failed.\n");
  return 1;
}

int query_prevent_shadow() { return 1; }
int query_open_module() { return 1; }
